'' System Version 2.001
'' OS_dspText Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/10/05 - Corrected windowColor, windowChar, screenColor, screenChar
''   2006/10/16 - Corrected Bsp routine for TV and lores VGA
''   2006/10/26 - Added constants for display type

'' This is the lores VGA windowed text driver for the Propeller OS.
'' It is based on tv_wtext.spin version 1.2 by Phil Pilgrim which
'' is in turn based on tv_text.spin which is  2006 by Parallax, Inc.
'' Modifications  2006 by Michael Green.  Permission is given to
'' include this or portions of this program in any program for the
'' Parallax Propeller processor as long as proper credit is given.

CON
  cols         = 32
  rows         = 15
  screensize   = cols * rows
  TV           = false
  loresVGA     = true
  hiresVGA     = false

  ZAP_COLOR    = 1 << 10 + $2FE
  ZAP_CHAR     = !ZAP_COLOR | $200 

  dsp_col       =   0   'current screen column
  dsp_row       =   1   'current screen row
  dsp_color     =   2   'current color
  dsp_states    =   3   'current state flags ("lazy" eol, insert)
  dsp_wleft     =   4   'current window left column
  dsp_wtop      =   5   'current window top row
  dsp_wcols     =   6   'current window number of columns
  dsp_wrows     =   7   'current window number of rows
  dsp_window    =   8   'currently selected window
  dsp_flag      =   9   'true if storing parameters
  dsp_pcnt      =  10   'number of parameter bytes so far
  dsp_pptr      =  11   'current pointer to parameter buffer
  dsp_cog       =  12   'cog number and bitmap for tv driver
  dsp_params    =  14   'parameter buffer for screen commands (6 x 1)
  dsp_windows   =  20   'dsp_col through dsp_wrows for each window (8 x 16)
  dsp_status    = 148   '0/1/2 = off/invisible/visible              read-only  (14 longs)
  dsp_enable    = 152   '0/non-0 = off/on                           write-only
  dsp_pins      = 156   '%pppttt = pins                             write-only
  dsp_mode      = 160   '%tihv = tile,interlace,hpol,vpol           write-only
  dsp_screen    = 164   'pointer to screen (words)                  write-only
  dsp_colors    = 168   'pointer to colors (longs)                  write-only 
  dsp_ht        = 172   'horizontal tiles                           write-only
  dsp_vt        = 176   'vertical tiles                             write-only
  dsp_hx        = 180   'horizontal tile expansion                  write-only 
  dsp_vx        = 184   'vertical tile expansion                    write-only 
  dsp_ho        = 188   'horizontal offset                          write-only  
  dsp_vo        = 192   'vertical offset                            write-only
  dsp_hd        = 196   'horizontal display ticks                   write-only
  dsp_hf        = 200   'horizontal front porch ticks               write-only
  dsp_hs        = 204   'horizontal sync ticks                      write-only
  dsp_hb        = 208   'horizontal back porch ticks                write-only
  dsp_vd        = 212   'vertical display lines                     write-only
  dsp_vf        = 216   'vertical front porch lines                 write-only
  dsp_vs        = 220   'vertical sync lines                        write-only
  dsp_vb        = 224   'vertical back porch lines                  write-only
  dsp_rate      = 228   'tick rate (Hz)                             write-only
  dsp_maxSize   = 232

  dsp_LazyEol   = %00000001                    ' dsp_status: at right end of window
  dsp_Insert    = %00000010                    ' dsp_status: insert mode (for editor)
  
'' Control codes.  For convenience, supplying a # cols and # rows of zero will use the maximum
'' allowable values (screen width - x, screen height - y).  Because of the limitations on colors
'' (foreground and background for each row only), the Color command will set the colors for a
'' specific row.  The colors do not scroll with the text.

  ClrWin      = $00                            ' clear window  
  Home        = $01                            ' home in window
  MoveXY      = $02                            ' move to X,Y in window (X and Y follow)
  CursLt      = $03                            ' cursor left
  CursRt      = $04                            ' cursor right
  CursUp      = $05                            ' cursor up
  CursDn      = $06                            ' cursor down
  Color       = $07                            ' select color C (0-15) (C follows)
  Bsp         = $08                            ' backspace
  Tab         = $09                            ' tab (8 spaces per)
  Lf          = $0A                            ' linefeed
  ClrEol      = $0B                            ' clear to end of line
  ClrEow      = $0C                            ' clear to end of active window
  Cr          = $0D                            ' carriage return
  MoveX       = $0E                            ' move to X in window (X follows)
  MoveY       = $0F                            ' move to Y in window (Y follows)
  SetWin      = $10                            ' set bounds of window (1-15,x,y,cols,rows)
  UseWin      = $11                            ' change to selected window (0-15) (W follows)
  SetColor    = $12                            ' change color (0-15) to FG, BG (C, FG, BG follow)
  Insert      = $14                            ' insert (move line to right one column)
  ScrollUp    = $15                            ' scroll up the current window
  ScrollDn    = $16                            ' scroll down the current window
  WinColor    = $17                            ' change all colors in active window (C follows)
  EscChr      = $1F                            ' (ESC) print next character C as-is (C follows)
  Delete      = $7F                            ' delete current character
  Alt00       = $FF                            ' changed to $00 by out (for use in string())
  ClrScr      = Alt00                          ' equivalent to ClrWin (for use in string())

  absX        = %0000                          ' X value is relative to screen coordinates             
  relX        = %0001                          ' X value is relative to current values
  winX        = %0010                          ' X value is relative to window coordinates
  absY        = %0000                          ' Y value is relative to screen coordinates
  relY        = %0100                          ' Y value is relative to current values
  winY        = %1000                          ' Y value is relative to window coordinates

  NoChgColor  = 16                             ' color # to indicate no change to current color
  UseWinColor = 17                             ' color # to cause use of current window color

OBJ
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  
PUB out(c) | i, j, k, p, f, x, y, t
'' Output a character.  This routine decodes the character stream, interpreting control codes.
'' These mostly follow the codes used with other text drivers with the exception of color since
'' the availability and details of color assignment vary so much from driver to driver.

  p := long[ldr#videoPtr]
  if c == $FF
    c~
  if f := byte[p+dsp_flag]
    byte[p+dsp_params][byte[p+dsp_pptr]++] := c
    if byte[p+dsp_pptr] == byte[p+dsp_pcnt]
      case f
        MoveXY:
          updateCursor(byte[p+dsp_params][0],c,winX|winY,false)
          byte[p+dsp_states] &= !dsp_LazyEol
        SetWin:
          i := (byte[p+dsp_params][0] & $0F) * 8
          if i > 0
            x := byte[p+dsp_params][1] <# (cols - 1)
            y := byte[p+dsp_params][2] <# (rows - 1)
            if (j := byte[p+dsp_params][3] <# (cols - x)) == 0
              j := cols - x
            if (k := c <# (rows - y)) == 0
              k := rows - y
            t := byte[p+dsp_color]
            long[p+dsp_windows+i][0] := 0 << 24 | t << 16 | y << 8 | x
            long[p+dsp_windows+i][1] := k << 24 | j << 16 | y << 8 | x
        UseWin:
          i := byte[p+dsp_window] * 8
          bytemove(p+dsp_windows+i, p+dsp_col, 4)
          i := (byte[p+dsp_window] := c & $0F) * 8
          bytemove(p+dsp_col, p+dsp_windows+i, 8)
        SetColor:
          i := byte[p+dsp_params][0] & $0F   ' color number
          j := byte[p+dsp_params][1]         ' foreground color
          t := long[ldr#colorPtr]            ' background color == c
          long[t][i << 1]     := j << 24 | c << 16 | j << 8 | c
          long[t][i << 1 + 1] := j << 24 | j << 16 | c << 8 | c
        WinColor:
          x := byte[p+dsp_wleft]
          y := byte[p+dsp_wtop]
          t := long[ldr#screenPtr]
          c := c & $0F
          repeat i from y to y + byte[p+dsp_wrows] - 1
            repeat j from x to x + byte[p+dsp_wcols] - 1
              k := i * cols + j
              word[t][k] := word[t][k] & ZAP_COLOR | c << 11
          byte[p+dsp_color] := c
        MoveX:
          updateCursor(c,0,winX|relY,false)
          byte[p+dsp_states] &= !dsp_LazyEol
        MoveY:
          updateCursor(0,c,relX|winY,false)
        EscChr:
          print(c)
        Color:
          byte[p+dsp_color] := c & $0F
      byte[p+dsp_flag]~
  else
    case c & $FF
      ClrWin:
        y := byte[p+dsp_wtop]
        t := byte[p+dsp_wrows]
        repeat i from y to y + t - 1
          fillwith(byte[p+dsp_wleft],i,byte[p+dsp_wcols]," ")
        updateCursor(0,0,winX|winY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      Home:     
        updateCursor(0,0,winX|winY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      MoveXY:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 2
      CursLt:
        updateCursor(-1,0,relX|relY,false)
        byte[p+dsp_states] &= !dsp_LazyEol
      CursRt:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          updateCursor(+1,0,relX|relY,true)
      CursUp:
        updateCursor(0,-1,relX|relY,false)
      CursDn:
        updateCursor(0,+1,relX|relY,false)
      SetWin:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 5
      Bsp:
        x := byte[p+dsp_col] - byte[p+dsp_wleft]
        y := byte[p+dsp_row] - byte[p+dsp_wtop]
        if x | y
          if x
            x--
          else
            x := byte[p+dsp_wcols] - 1
            y--
          updateCursor(x,y,winX|winY,false)
          setScreenChar(byte[p+dsp_col],byte[p+dsp_row]," ",UseWinColor)
      Tab:
        repeat
          print(" ")
        while (byte[p+dsp_col] - byte[p+dsp_wleft]) & 7
      Lf:
        if (byte[p+dsp_row] - byte[p+dsp_wtop]) == byte[p+dsp_wrows] - 1
          x := byte[p+dsp_col]
          y := byte[p+dsp_row]
          newline
          updateCursor(x,y,absX|absY,false)
        else
          updateCursor(0,+1,relX|relY,false)
      ClrEol:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
      ClrEow:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
        y := byte[p+dsp_row]
        t := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
        repeat i from y + 1 to t
          fillwith(byte[p+dsp_wleft],i,byte[p+dsp_wcols]," ")
      SetColor:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 3
      Cr:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          cleartoeol
        newline
      MoveX, MoveY, UseWin, EscChr, WinColor, Color:
        byte[p+dsp_flag] := c
        byte[p+dsp_pptr]~
        byte[p+dsp_pcnt] := 1
      Insert:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          t := long[ldr#screenPtr]
          i := byte[p+dsp_row] * cols
          k := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
          repeat j from k to byte[p+dsp_col] + 1
            word[t][i + j] := word[t][i + j - 1]
          setScreenChar(byte[p+dsp_col],byte[p+dsp_row]," ",UseWinColor)
      Delete:
        if (byte[p+dsp_states] & dsp_LazyEol) == 0
          t := long[ldr#screenPtr]
          i := byte[p+dsp_row] * cols
          k := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
          repeat j from byte[p+dsp_col] to k - 1
            word[t][i + j] := word[t][i + j + 1]
          setScreenChar(k,byte[p+dsp_row]," ",UseWinColor)
      ScrollUp:
        scrollUpWin
      ScrollDn:
        scrollDownWin
      other:
        print(c)

PUB str(stringptr)
'' Print a zero-terminated string
  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i
'' Print a decimal number
  if value < 0
    -value
    out("-")
  i := 1_000_000_000
  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB decFixed(value,width) | i, p, f, a, b, c
'' Print a decimal number of fixed width
  p := @a
  f~
  if value < 0
    -value
    byte[p++] := "-"
  i := 1_000_000_000
  repeat 10
    if value => i
      byte[p++] := value / i + "0"
      value //= i
      f~~
    elseif f or i == 1
      byte[p++] := "0"
    i /= 10
  byte[p]~
  i := strsize(@a)
  if i < width
    repeat width - i
      out(" ")
  str(@a)
  
PUB hex(value, digits)
'' Print a hexadecimal number
  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number
  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

PUB screenChar(x, y) | p, s, c
'' Return the character at absolute screen coordinates x,y
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return " "
  s := long[ldr#screenPtr]
  c := word[s][y * cols + x]
  return (c & $FE) | ((c >> 10) & 1)

PUB screenColor(x, y) | p, s
'' Return the color of the character at absolute screen coordinates x,y
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return NoChgColor
  s := long[ldr#screenPtr]
  return word[s][y * cols + x] >> 11

PUB setScreenChar(x, y, c, z) | p, s
'' Set the character at absolute screen coordinates x,y to c with color z
'' If z == NoChgColor, the color is not changed.
'' If z == UseWinColor, the current window color is used.
  p := long[ldr#videoPtr]
  if x => cols or y => rows
    return
  s := long[ldr#screenPtr]
  case z
    NoChgColor:
      z := word[s][y * cols + x] >> 11
    UseWinColor:
      z := byte[p+dsp_color]
  word[s][y * cols + x] := (z << 1 + c & 1) << 10 + $200 + c & $FE

PUB windowChar(x, y) | p
'' Return the character at window relative coordinates x,y
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return " "
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  return screenChar(x,y)

PUB windowColor(x, y) | p
'' Return the color of the character at window relative coordinates x,y
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return NoChgColor
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  return screenColor(x,y)

PUB setWindowChar(x, y, c, z) | p
'' Set the character at window relative coordinates x,y to c with color z
'' If z == NoChgColor, the color is not changed.
'' If z == UseWinColor, the current window color is used.
  p := long[ldr#videoPtr]
  if x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]
    return " "
  x += byte[p+dsp_wleft]
  y += byte[p+dsp_wtop]
  setScreenChar(x,y,c,z)

PUB windowAtEnd(x, y) | p
'' Return true if window relative coordinates are past boundary
  p := long[ldr#videoPtr]
  return x => byte[p+dsp_wcols] or y => byte[p+dsp_wrows]

PUB windowCol | p
'' Return the current window relative x coordinate (column)
  p := long[ldr#videoPtr]
  return byte[p+dsp_col] - byte[p+dsp_wleft]

PUB windowRow | p
'' Return the current window relative y coordinate (row)
  p := long[ldr#videoPtr]
  return byte[p+dsp_row] - byte[p+dsp_wtop]

PUB out2(a, b)
'' Output a 2 character sequence
  out(a)
  out(b)

PUB out3(a, b, c)
'' Output a 3 character sequence
  out(a)
  out(b)
  out(c)

PUB out4(a, b, c, d)
'' Output a 4 character sequence
  out(a)
  out(b)
  out(c)
  out(d)

PUB out5(a, b, c, d, e)
'' Output a 5 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)

PUB out6(a, b, c, d, e, f)
'' Output a 6 character sequence
  out(a)
  out(b)
  out(c)
  out(d)
  out(e)
  out(f)

PRI print(c) | p
  p := long[ldr#videoPtr]
  if (byte[p+dsp_states] & dsp_LazyEol) <> 0
    newline
  setScreenChar(byte[p+dsp_col],byte[p+dsp_row],c,UseWinColor)
  updateCursor(+1,0,relX|relY,true)

PRI updateCursor(x,y,mode,chkEol) | p, ax, ay, wx, wy, t
  p := long[ldr#videoPtr]
  ax := byte[p+dsp_wleft]
  ay := byte[p+dsp_wtop]
  wx := ax + byte[p+dsp_wcols] - 1
  wy := ay + byte[p+dsp_wrows] - 1
  case mode & %0011
    relX:
      x += byte[p+dsp_col]
      x &= x & $80000000 == 0
    winX:
      x += ax
  case (mode & %1100) >> 2
    relY >> 2:
      y += byte[p+dsp_row]
      y &= y & $80000000 == 0
    winY >> 2:
      y += ay
  if chkEol
    if x > wx or y > wy
      byte[p+dsp_states] |= dsp_LazyEol
    else
      byte[p+dsp_states] &= !dsp_LazyEol
  byte[p+dsp_col] := x := x #> ax <# wx
  byte[p+dsp_row] := y := y #> ay <# wy

PRI newline | p
  p := long[ldr#videoPtr]
  updateCursor(0,+1,winX|relY,true)
  if (byte[p+dsp_states] & dsp_LazyEol) <> 0
    scrollUpWin
  byte[p+dsp_states] &= !dsp_LazyEol

PRI scrollUpWin | i, k, p, e
  p := long[ldr#videoPtr]
  if byte[p+dsp_wrows] > 1
    e := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
    repeat i from byte[p+dsp_wtop] to e - 1
      moveline(byte[p+dsp_wleft],i,byte[p+dsp_wcols],i+1)
    k := byte[p+dsp_col]
    updateCursor(0,e,winX|absY,false)
    cleartoeol
    updateCursor(k,0,absX|relY,false)

PRI scrollDownWin | i, k, p, e
  p := long[ldr#videoPtr]
  if byte[p+dsp_wrows] > 1
    e := byte[p+dsp_wtop] + byte[p+dsp_wrows] - 1
    repeat i from e to byte[p+dsp_wtop] + 1
      moveline(byte[p+dsp_wleft],i,byte[p+dsp_wcols],i-1)
    k := byte[p+dsp_col]
    updateCursor(0,0,winX|winY,false)
    cleartoeol
    updateCursor(k,0,absX|winY,false)

PRI cleartoeol | p, x, w
 p := long[ldr#videoPtr]
 x := byte[p+dsp_col]
 w := byte[p+dsp_wleft] + byte[p+dsp_wcols] - 1
 fillwith(x,byte[p+dsp_row],w-x," ")
       
PRI fillwith(x, y, w, c) | p, i, s, clr
  p := long[ldr#videoPtr]
  s := long[ldr#screenPtr]
  clr := byte[p+dsp_color]
  repeat i from x to x + w - 1
    word[s][y * cols + i] := (clr << 1 + c & 1) << 10 + $200 + c & $FE

PRI moveline(x, y, w, f) | i, s
  s := long[ldr#screenPtr]
  repeat i from x to x + w - 1
    word[s][y * cols + i] := word[s][f * cols + i]
