'' System Version 2.001
'' OS_mainProgram Changes:
''
''   2006/09/24 - Initialization routine moved to its own object.  This object
''                is now the command interpreter and some support routines and
''                is the main (top) object in the Operating System.  Test commands
''                added for the uALFAT/USBwiz for MMC/SD memory card access.
''   2006/10/02 - Updated some comments
''   2006/10/03 - Improved screen formatting on small screens

'' This is the command interpreter for the Propeller "Operating System".
'' This set of objects is intended to provide a basic set of I/O including TV/VGA
'' text , a PS/2 keyboard, and management of one or more I2C memory busses, each
'' containing up to 256K of EEPROM.  One is always the boot I2C bus on I/O pins
'' 28 and 29.  Others can be added on any pair of I/O pins.  Included is a
'' SPIN loader that can load a normal SPIN program from one of the EEPROMs
'' known to the system, from any location in the EEPROM.  The size of the program
'' is stored with the program by the IDE, but this can be overridden.  The I/O
'' drivers are intended to run continuously in 3 or 4 COGs (1 or 2 for display,
'' 1 for the keyboard, and 1 for I2C).  Their work areas (including buffers) are
'' kept at the end of RAM memory so they should not be overwritten when a new
'' program is loaded.  The new program, if it uses these display text, keyboard,
'' and loader objects, should be able to seamlessly use the display and keyboard
'' and transfer control further to other programs and back to this main routine
'' for further interaction.  The loader will stop any COGs not known to it prior
'' to loading a new progam.  Programs can register COGs intended to "persist"
'' between programs.  These must be running assembly language and use a work area
'' that won't be overlaid.  There are methods provided by the loader to do simple
'' work area allocation.

'' There are several objects that are display dependent, OS_initialization,
'' OS_loader, OS_dspText, and OS_dspTextInit.  OS_initialization contains the
'' display I/O pin definitions and also contains the definitions for the PS/2
'' keyboard pins.  OS_loader contains display dependent memory allocation
'' information.  OS_dspText and OS_dspTextInit contain display specific code.
'' There are 3 sets of objects: one for a TV text display (40 x 13), one for
'' a low resolution VGA text display (32 x 15), and one for a high resolution
'' VGA text display.  The high resolution VGA text display can be compiled for
'' 3 different text resolutions: 80 x 40 (640 x 480), 100 x 50 (800 x 600), and
'' 128 x 64 (1024 x 768).  These are defined in OS_dspText as blocks of constants
'' two of which are commented out with the default being 80 x 40.  Simply comment
'' out the active one and uncomment the one for the desired resolution then
'' recompile.  This cannot be used with the 512 x 384 VGA graphics/text driver.
'' It uses too much memory for its bitmap buffer.  This also cannot currently
'' be used with any of the graphics (TV or low resolution VGA) drivers.  A program
'' that uses these drivers can be run by this OS, but the OS drivers have to be
'' shut down (stopped) and the new program has to supply all of its own I/O, then
'' reload the OS when it's done.  The new program can use OS_loaderInit to do so.

'' All routines (except drivers where noted) are (c) 2006 by Michael Green.
'' Some I/O drivers are (c) by Parallax, some with modifications by others.
'' See the notes at the beginning of each source file for specifics.
'' Permission is given to include this or portions of this program in any program
'' for the Parallax Propeller processor as long as proper credit is given.

CON
  _clkmode    = xtal1 + pll16x
  _xinfreq    = 5_000_000
  
OBJ
  init    : "OS_initialization"                ' OS Initialization routines and definitions
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  dsp     : "OS_dspText"                       ' Display text driver modified for OS use
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  files   : "OS_fileSystem"                    ' File system support
  edit    : "OS_windowEditor"                  ' Screen and file editor
  load    : "OS_eepromLoader"                  ' Downloads from Propeller Tool to EEPROM
  USBwiz  : "USBwiz_Interface"                 ' Test interface for USBwiz OEM

CON
  _free = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4  ' # longs needed
                                               
  #1, processProbe, processInfo,  processEdit  ' For command post-processing
      processHelp,  processRun,   processList
      processCopy,  processDump,  processDelete
      processLoad
      processTestList,    processTestChange
      processTestClose,   processTestFlush
      processTestRead,    processTestDelete
      processTestConnect, processTestOpen
      processTestVersion, processTestCopy
      processTestReset

VAR
  long dumpPin, dumpAddr, dumpSize
  long fileAddr, fileAddrTo, fileChange
  long fileName[3], fileNameTo[3]

PUB processCommands | t, y, i, j, k, z, b0, b1, b2, b3, b4, b5, b6, b7
'' Initialize the operating system, then repeatedly request a one line
'' command from the user (in a single line window), then execute it in a
'' full screen window.
  init.start(string("Propeller OS - Version 2.001.04",dsp#Cr))
  repeat
    y := dsp.windowRow
    t := \enterDoCommand(y)
    if t < 0
      dsp.out2(dsp#UseWin,0)
      dsp.out4(dsp#MoveXY,dsp#cols,y,dsp#Cr)
      dsp.str(string("*** "))
      dsp.str(-t)
      dsp.out(dsp#Cr)
    else
      dsp.out2(dsp#UseWin,0)
      dsp.out4(dsp#MoveXY,dsp#cols,y,dsp#Cr)
      case t
        processProbe:
          if dumpPin < (ldr#maxPinPairs << 1)
            if (\files.updateBitmapEEPROMs(dumpPin) & $4000) == 0
              dsp.str(string("*** no boot eeprom?",dsp#Cr))
              return
          repeat i from 0 to (ldr#maxPinPairs<<1)-2 step 2
            z~
            y := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
            if y <> 0
              repeat j from 0 to 15
                if (|<j & y) <> 0
                  z += 32
              dsp.str(string("    Pin "))
              dsp.decFixed(i,2)
              dsp.str(string("  "))
              dsp.decFixed(z,3)
              dsp.str(string("K",dsp#Cr))
        processInfo:
          dsp.str(string("    pbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_pbase,@t,4)
          dsp.hex(t,4)
          dsp.str(string(", vbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_vbase,@t,4)
          dsp.hex(t,4)
          dsp.str(string(", dbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_dbase,@t,4)
          dsp.hex(t,4)
          dsp.str(string(dsp#Cr,"    pcurr="))
          ldr.readEEPROM(fileAddr+ldr#spin_pcurr,@t,4)
          dsp.hex(t,4)
          dsp.str(string(", dcurr="))
          ldr.readEEPROM(fileAddr+ldr#spin_dcurr,@t,4)
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processEdit:
          editWindow
        processHelp:
'' Copy 32K (0-$7FEF) from one file to another.  Create a file if needed.
          dsp.str(string(" copy <fileFrom> <fileTo>",dsp#Cr))
'' Delete the specified file
          dsp.str(string(" delete <file>",dsp#Cr))
'' Display the specified EEPROM or HUB RAM locations
          dsp.str(string(" dump {<pin>} <addr> <len>",dsp#Cr))
'' Start a simple screen editor.  Quit using the App Key (to the left of Ins)
          dsp.str(string(" edit",dsp#Cr))
'' Display a list of accepted commands
          dsp.str(string(" help",dsp#Cr))
'' Show SPIN related pointers for the program in the specified file
          dsp.str(string(" info <file>",dsp#Cr))
'' List files available including their location and option bits
          dsp.str(string(" list",dsp#Cr))
'' Download Propeller Tool to EEPROM
          dsp.str(string(" load <pin><page><Rst><Rx><Tx>",dsp#Cr))
'' Optionally probe the possible I2C buss on the specified pin pair for EEPROMs
'' List all known I2C busses and the total amount of EEPROM space on each
          dsp.str(string(" probe {<pin>}",dsp#Cr))
'' Execute the SPIN program contained in the specified file
          dsp.str(string(" run <file>",dsp#Cr))
        processList:
          files.fileList
        processDump:
          files.dumpMemory(dumpPin,dumpAddr,dumpSize)
        processCopy:
          if fileChange
            dsp.str(string("    Overwrite? (Y/n) "))
            repeat
              t := key.getkey
              if t == dsp#Cr
                t := "y"
            until t == "Y" or t == "y" or t == "N" or t == "n"
            dsp.out(t)
          else
            t := "y"
            dsp.str(string("Created"))
          if t == "Y" or t == "y"
            dsp.out(" ")
            j := dsp.windowCol
            repeat i from 0 to $7FE0 step 32
              dsp.out2(dsp#MoveX,j)
              dsp.decFixed(i,5)
              if i == $7FE0
                k := 16
              else
                k := 32
              t := cnt
              repeat until ldr.readEEPROM(fileAddr+i,@b0,k)
                if (cnt - t) > (clkfreq / 200)
                  abort -string("EEPROM read timeout")
              t := cnt
              repeat until ldr.writeEEPROM(fileAddrTo+i,@b0,k)
                if (cnt - t) > (clkfreq / 200)
                  abort -string("EEPROM write timeout")
            dsp.out2(dsp#MoveX,j)
            dsp.str(string("Saved"))
          dsp.out(dsp#Cr)
        processRun:
          dsp.str(string("    Loading ... ",dsp#Cr))
          repeat i from 0 to 7  ' Stop any cogs not registered
            if (word[long[ldr#systemPtr]+ldr#sys_cog] & |<i) == 0
              cogstop(i)
          ldr.bootEEPROM(fileAddr)
          dsp.str(string("    Error during loading",dsp#Cr))
        processDelete:
          dsp.str(string("    Are you sure? (Y/n) "))
          repeat
            t := key.getkey
            if t == dsp#Cr
              t := "y"
          until t == "Y" or t == "y" or t == "N" or t == "n"
          dsp.out(t)
          if t == "Y" or t == "y"
            files.markDeleted(@fileName,fileAddr)
            dsp.str(string(" - Done"))
          dsp.out(dsp#Cr)
        processLoad:
          dsp.str(string("    Address = "))
          dsp.hex(fileAddr,6)
          dsp.str(string(dsp#Cr,"    Rst,Rx,Tx = "))
          dsp.dec(dumpPin)
          dsp.out(",")
          dsp.dec(dumpPin+1)
          dsp.out(",")
          dsp.dec(dumpPin+2)
          dsp.out(dsp#Cr)
          load.start(dumpPin,dumpPin+1,dumpPin+2,fileAddr)
        processTestList:
          t := \USBwiz.listDirectory(500)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestOpen:
          t := \USBwiz.openFile(-500,@fileName,"R")
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestChange:
          t := \USBwiz.changeDir(-500,@fileName)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestClose:
          t := \USBwiz.closeFile(-500)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestFlush:
          t := \USBwiz.flushFile(-500)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestRead:
          t := \USBwiz.readFile(-500,@b0,32)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
          if t == 0
            files.dumpMemory(ldr#maxPinPairs << 1,@b0,32)
        processTestDelete:
          t := \USBwiz.deleteFile(-500,@fileName)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestConnect:
          t := \USBwiz.connectCard(-500)
          dsp.str(string("    Result = "))
          dsp.hex(t,4)
          dsp.out(dsp#Cr)
        processTestVersion:
          t := \USBwiz.getVersion(-500)
        processTestCopy:
          z := true
          repeat
            case t := key.key
              13, " ".."~":
                USBwiz.sendChar(-500,t)
                z := true
              $CC:
                quit
            case t := \USBwiz.recvChar(500)
              13, " ".."~":
                dsp.out(t)
              other:
                if z~
                  dsp.out("<")
                  dsp.dec(t)
                  dsp.out(">")
        processTestReset:
          t := fileChange
          \USBwiz.start(t>>24, (t>>16)&$1F, (t>>8)&$1F, t&$1F)
        
PUB editWindow | t
'' Set up a simple screen editor with numbered blank lines.  Use the "App" key
'' to exit the editor (this is left of the "Ins" key on the Parallax keyboard).
  dsp.out6(dsp#SetWin,1,5,1,dsp#cols-5,dsp#rows-2)
  dsp.out6(dsp#SetWin,2,0,1,5,dsp#rows-2)
  dsp.out5(dsp#UseWin,2,dsp#WinColor,0,dsp#ClrScr)
  repeat t from 1 to dsp#rows - 2
    dsp.out3(dsp#MoveXY,0,t-1)
    dsp.decFixed(t,4)
  dsp.out3(dsp#UseWin,1,dsp#ClrScr)
  repeat
    t := edit.windowEditor
  until t == $CC
  dsp.out3(dsp#UseWin,0,dsp#ClrScr)

PUB enterDoCommand(y) | t, x, a, b, c
'' Set up a window at the line specified and accept a command
'' line terminated by a Cr.  Scan for a command word and
'' command specific parameters, either words or numbers.
'' If the command is known and the parameters are valid,
'' execute the command.  Abort with a negative error string
'' address or return a positive command specific value.
  dsp.out6(dsp#SetWin,1,1,y,dsp#cols-1,1)
  dsp.out6(dsp#UseWin,0,dsp#MoveXY,0,y,">")
  dsp.out5(dsp#UseWin,1,dsp#WinColor,0,dsp#ClrScr)
  repeat
    t := edit.windowEditor
  until t == dsp#Cr
  x := \getTextWord(0,@a,4,true," ")
  if x < 0
    abort -string("no command given")
  case a
    $59504F43:  ' copy
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <fileFrom>")
      x := \getTextWord(x,@fileNameTo,12,true," ")
      if x < 0
        abort -string("missing <fileTo>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<fileFrom> not found")
      fileChange := true
      fileAddrTo := \files.searchFile(@fileNameTo)
      if fileAddrTo < 0
        fileAddrTo := \files.searchFile(0)
        if fileAddrTo < 0
          abort -string("can't create <fileTo>")
        files.markNewName(@fileNameTo,fileAddrTo,0)
        fileChange := false
      return processCopy
    $424F5250:  ' probe
      x := \getNumber(x,@dumpPin)
      if x < 0
        dumpPin := ldr#maxPinPairs << 1
        return processProbe
      if (dumpPin < (ldr#maxPinPairs << 1)) and (dumpPin & 1 == 0)
        return processProbe
      else
        abort -string("pin number odd or does not exist")
    $204E5552:  ' run
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processRun
    $54494445:  ' edit
      return processEdit
    $4F464E49:  ' info
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processInfo
    $504C4548:  ' help
      return processHelp
    $5453494C:  ' list
      return processList
    $454C4544:  ' delete
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      if fileAddr == (ldr#i2cBootSCL << 18)
        abort -string("can't delete boot EEPROM")
      return processDelete
    $504D5544:  ' dump
      x := \getNumber(x,@a)
      if x < 0
        abort -string("missing <pin> or <addr>")
      x := \getNumber(x,@b)
      if x < 0
        abort -string("missing <addr> or <len>")
      x := \getNumber(x,@c)
      if x < 0
        if (a > $7FFFF) or (b => 1024)
          abort -string("invalid number")
        dumpPin := ldr#maxPinPairs << 1
        dumpAddr := a
        dumpSize := b
        return processDump
      else
        if (a => (ldr#maxPinPairs << 1)) or (a & 1) or (b > $7FFFF) or (c => 1024)
          abort -string("invalid number")
        dumpPin := a
        dumpAddr := b
        dumpSize := c
        return processDump
    $44414F4C:  ' load
      x := \getNumber(x,@b)
      if x < 0
        abort -string("missing <pin>")
      x := \getNumber(x,@c)
      if x < 0
        abort -string("missing <page>")
      if (b => (ldr#maxPinPairs << 1)) or (b & 1) or (c => 16)
        abort -string("invalid number")
      x := \getNumber(x,@dumpPin)
      if x < 0
        abort -string("missing <Rst/Rx/Tx>")
      fileAddr := (b << 18) | (c << 15)
      return processLoad
    $54534554:  ' test
      x := \getTextWord(x,@a,4,true," ")
      if x < 0
        abort -string("missing <function>")
      case a
        $5453494C:  ' list
          return processTestList
        $4E45504F:  ' open <file>
          if (x := \getTextWord(x,@fileName,12,true," ")) < 0
            abort -string("missing <file>")
          return processTestOpen
        $4E414843:  ' change <dir>
          if (x := \getTextWord(x,@fileName,12,true," ")) < 0
            abort -string("missing <dir>")
          return processTestChange
        $534F4C43:  ' close
          return processTestClose
        $53554C46:  ' flush
          return processTestFlush
        $44414552:  ' read
          return processTestRead
        $454C4544:  ' delete <file>
          if (x := \getTextWord(x,@fileName,12,true," ")) < 0
            abort -string("missing <file>")
          return processTestDelete
        $4E4E4F43:  ' connect
          return processTestConnect
        $53524556:  ' version
          return processTestVersion
        $59504F43:  ' copy
          return processTestCopy
        $45534552:  ' reset <Rdy><Bsy><SCL><SDA>
          x := \getNumber(x,@a)
          if x < 0
            abort -string("missing <Rdy/Bsy/SCL/SDA>")
          x := \getNumber(x,@b)
          if x < 0
            abort -string("missing <Rdy/Bsy/SCL/SDA>")
          x := \getNumber(x,@c)
          if x < 0
            abort -string("missing <Rdy/Bsy/SCL/SDA>")
          x := \getNumber(x,@t)
          if x < 0
            abort -string("missing <Rdy/Bsy/SCL/SDA>")
          fileChange := a << 24 | b << 16 | c << 8 | t
          return processTestReset
        other:
          abort -string("unknown function")
    other:
      abort -string("unknown command")

PRI skipBlanks(x)
'' Skip blanks until at right border of window and return x
'' of first non-blank or abort with a negative error string
'' address.
  repeat
    if dsp.windowAtEnd(x,0)
      abort -string("missing non-blank")
  until dsp.windowChar(x++,0) <> " "
  return x - 1

PRI getTextWord(x,textAddr,maxSize,uppercase,clearFirst) | c
'' Skip leading blanks, then copy up to maxSize non-blank
'' characters into the area at textAddr.  Return the
'' position of the next blank.  If uppercase is true,
'' store any lower case alphabetic characters as upper
'' case.  If clearFirst is not -1 (true), clear the
'' textAddr area to the value given.  If an error occurs,
'' abort with a negative error string address
  if clearFirst <> -1
    repeat c from 1 to maxSize
      byte[textAddr + c - 1] := clearFirst
  x := skipBlanks(x)
  repeat while (c := dsp.windowChar(x++,0)) <> " "
    if maxSize > 0
      if uppercase and c => "a" and c =< "z"
        c := c & !%0100000
      byte[textAddr++] := c
      maxSize--
  return x - 1

PRI getNumber(x,numberAddr) | negative, c, t
'' Skip leading blanks, then convert a decimal number
'' with an optional sign ("+"/"-") into binary to be
'' stored in the long at numberAddr.  Return the
'' position of the next blank.  If an error occurs,
'' abort with a negative error string address.
'' Currently, there's no checking for 32 bit overflow.
  x := skipBlanks(x)
  c := dsp.windowChar(x++,0)
  negative := c == "-"
  if negative or c == "+"
    c := dsp.windowChar(x++,0)
  if c == "$"
    c := dsp.windowChar(x++,0)
    t := hexDigit(c)
    if t < 0
      abort -string("invalid hex character")
    long[numberAddr] := t
    c := dsp.windowChar(x++,0)
    repeat until (t := hexDigit(c)) < 0
      long[numberAddr] := long[numberAddr] << 4 | t
      c := dsp.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if c == "%"
    c := dsp.windowChar(x++,0)
    if not (c == "0" or c == "1")
      abort -string("invalid hex character")
    long[numberAddr] := c - "0"
    c := dsp.windowChar(x++,0)
    repeat while c == "0" or c == "1"
      long[numberAddr] := long[numberAddr] << 1 | (c - "0")
      c := dsp.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if not (c => "0" and c=< "9")
    abort -string("invalid numeric character")
  long[numberAddr] := c - "0"
  c := dsp.windowChar(x++,0)
  repeat while c => "0" and c =< "9"
    long[numberAddr] := long[numberAddr] * 10 + c - "0"
    c := dsp.windowChar(x++,0)
  if negative
    long[numberAddr] := -long[numberAddr]
  return x - 1

PRI hexDigit(c)
'' Convert a hexadecimal character to the corresponding value or -1 if invalid.
  if c => "0" and c =< "9"
    return c - "0"
  if c => "A" and c =< "F"
    return c - "A" + 10
  if c => "a" and c =< "f"
    return c - "a" + 10
  return -1
  