'' System Version 2.001
'' USBwiz_Interface Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/10/10 - Revise error code return conventions
''   2006/10/14 - Supply the pin numbers in the start routine
''   2006/10/15 - Improved reset routine, added set clock

'' This object provides an I2C interface to the GHI Electronics
'' USBwiz-OEM MMC/SD card and USB interface processor.

CON
  USBwiz_EndDir = $4D                               ' End of directory information
  USBwiz_NDT    = $FF                               ' I2C No Data
  USBwiz_HDT    = $FE                               ' I2C Half Data

  USBwiz_Addr   = %10100100                         ' I2C Device Address

  ErrCode       = $0100                             ' Prefix for USBwiz errors  
  I2CIOErr      = $0222                             ' I2C I/O Error
  MiscErr       = $0333                             ' Miscellaneous error
  TimeOut       = $0444                             ' Timeout error

OBJ
  dsp : "OS_dspText"                                ' OS display driver interface
  ldr : "OS_loader"                                 ' OS I2C driver interface

VAR
  byte USBwiz_SDA, USBwiz_SCL, USBwiz_Bsy, USBwiz_Rdy
  
PUB start(Rdy, Bsy, SCL, SDA) | t                   '' Set up the I/O pins
  USBwiz_Rdy := Rdy
  USBwiz_Bsy := Bsy
  USBwiz_SCL := SCL
  USBwiz_SDA := SDA
  dira[USBwiz_Rdy] ~                                ' Set pins as inputs
  dira[USBwiz_Bsy] ~
  dira[USBwiz_SCL] ~
  dira[USBwiz_SDA] ~
  waitcnt(clkfreq / 100 + cnt)                      ' Wait 10ms for things to stabilize
  \sendChar(250,13)                                 ' Try to get the USBwiz to a known state
  repeat
    t := \recvChar(-250)                             ' Clear out any input chars
  until t == TimeOut or t == I2CIOErr or t == 0
  if t == TimeOut
    dsp.str(string("Timed Out",13))
  elseif t == I2CIOErr
    dsp.str(string("I2C Error",13))
  \sendStr(-250,string("RT OK",13))                 ' Do a software reset
  waitcnt(clkfreq / 100 + cnt)                      ' Wait 10ms for things to stabilize
  repeat
    t := \recvChar(-250)                            ' Clear out the identification string
  until t == TimeOut or t == I2CIOErr or t == 0
  if t == TimeOut
    dsp.str(string("Timed Out",13))
  elseif t == I2CIOErr
    dsp.str(string("I2C Error",13))

PUB stop                                            '' Make sure the pins are left as inputs
  dira[USBwiz_Rdy] ~
  dira[USBwiz_Bsy] ~
  dira[USBwiz_SCL] ~
  dira[USBwiz_SDA] ~

PUB setClock(ms,yr,mon,day,hr,mm,ss) | time         '' Set the USBwiz real time clock
  time := yr - 1980                                 '  Year   1980 to 2107
  time := time << 4 | mon                           '  Month  1 to 12
  time := time << 5 | day                           '  Day    1 to 31
  time := time << 5 | hr                            '  Hour   0 to 23
  time := time << 6 | mm                            '  Minute 0 to 59
  time := time << 5 | ss >> 1                       '  Second 0 to 59 (even only)
  sendStr(ms,string("ST "))
  sendHex(ms,time,8)
  sendChar(ms,13)
  checkErrorCode(ms,0)
  sendStr(ms,string("GT F",13))                     '  Verify what was entered
  return findErrorCode(ms)
  
PUB getVersion(ms)                                  '' Display the version information
  sendStr(ms,string("VR",13))
  return findErrorCode(ms)

PUB resetUSBwiz(ms)                                 '' Reset the USBwiz and display version
  sendStr(ms,string("RT OK",13))                    '  both firmware and firmware loader
  return findErrorCode(ms)
  
PUB connectCard(ms)                                 '' Prepare to access a storage device
  sendStr(ms,string("FM S",13))                     ' Mount file system (SD/MMC card)
  checkErrorCode(ms,0)

PUB listDirectory(ms) | d, a, s, n0, n1, n2, n3     '' Display the current directory
  sendStr(ms,string("IL",13))                       ' Initialize the directory listing
  checkErrorCode(ms,0)
  dsp.str(string("File Name    Attr     File Size",dsp#Cr))
  repeat
    sendStr(ms,string("NF",13))                     ' Get the next file entry
    if checkErrorCode(ms,USBwiz_EndDir)
      quit
    getFileName(ms,@n0)
    checkNext(ms," ")
    a := getHex(ms,2)                               ' Get file attribute
    checkNext(ms," ")
    s := getHex(ms,8)                               ' Get file size
    checkNext(ms,13)
    checkErrorCode(ms,0)
    dsp.str(@n0)                                    ' Display the information
    repeat 13 - strsize(@n0)
      dsp.out(" ")
    dsp.bin(a,8)
    dsp.out(" ")
    dsp.dec(s)
    dsp.str(string(" Bytes",dsp#Cr))
  sendStr(ms,string("MS",13))                       ' Get media statistics
  checkErrorCode(ms,0)
  a := getHexLong(ms)                               ' Get media size
  checkNext(ms," ")
  s := getHexLong(ms)                               ' Get free size
  checkNext(ms,13)
  checkErrorCode(ms,0)
  dsp.dec(a)
  dsp.str(string(" Total Size, "))
  dsp.dec(s)
  dsp.str(string(" Bytes Free",dsp#Cr))

PUB changeDir(ms, s)                                '' Change to specified subdirectory
  sendStr(ms,string("CD "))
  sendStr(ms,s)
  sendChar(ms,13)
  checkErrorCode(ms,0)

PUB openFile(ms, s, mode)                           '' Open specified file in the specified mode
  sendStr(ms,string("OF 0"))                        ' Send command and use handle zero
  sendChar(ms,mode)                                 ' Send file mode ("R", "W", "A")
  sendChar(ms,">")
  sendStr(ms,s)                                     ' Send file name
  sendChar(ms,13)
  checkErrorCode(ms,0)

PUB closeFile(ms)                                   '' Close current file
  sendStr(ms,string("CF 0",13))                     ' Send command and use handle zero
  checkErrorCode(ms,0)
  
PUB flushFile(ms)                                   '' Flush current file
  sendStr(ms,string("FF 0",13))                     ' Send command and use handle zero
  checkErrorCode(ms,0)
  
PUB readFile(ms, bufAdr, count) | t                 '' Read specified bytes from file
  sendStr(ms,string("RF 0",$7F,">"))
  sendHex(ms,count,8)
  sendChar(ms,13)
  checkErrorCode(ms,0)
  repeat count
    byte[bufAdr++] := recvChar(ms)
  if getHexLong(ms) <> count
    abort MiscErr
  checkNext(ms,13)
  checkErrorCode(ms,0)

PUB writeFile(ms, bufAdr, count) | t                 '' Write specified bytes to a file
  sendStr(ms,string("WF 0>"))
  sendHex(ms,count,8)
  sendChar(ms,13)
  checkErrorCode(ms,0)
  repeat count
    sendChar(ms,byte[bufAdr++])
  if getHexLong(ms) <> count
    abort MiscErr
  checkNext(ms,13)
  checkErrorCode(ms,0)

PUB deleteFile(ms, s)                               '' Delete specified file (not directory)
  sendStr(ms,string("DF "))
  sendStr(ms,s)
  sendChar(ms,13)
  checkErrorCode(ms,0)

PUB getFileName(ms, p) | i, name, name2, ext        '' Receive a fixed width 8.3 file name
  getTextField(ms,@name,8)                          '  Suppress trailing blanks and make
  checkNext(ms,".")                                 '   into a single zero-terminated string
  getTextField(ms,@ext,3)                           '   which may be as long as 13 bytes
  i := 0
  repeat while i < 8 and byte[@name][i] <> " "
    byte[p++] := byte[@name][i++]
  byte[p++] := "."
  i := 0
  repeat while i < 3 and byte[@ext][i] <> " "
    byte[p++] := byte[@ext][i++]
  byte[p] := 0

PUB getHexLong(ms)                                  '' Receive a hex long prefixed by "$"
  checkNext(ms,"$")
  return getHex(ms,8)

PUB getHexWord(ms)                                  '' Receive a hex word prefixed by "$"
  checkNext(ms,"$")
  return getHex(ms,4)

PUB getHexByte(ms)                                  '' Receive a hex byte prefixed by "$"
  checkNext(ms,"$")
  return getHex(ms,2)

PUB getTextField(ms, bufAddr, size) | c             '' Receive a fixed size text field
  repeat size
    byte[bufAddr++] := recvChar(ms)

PUB findErrorCode(ms) : err | c                     '' Ignore all up until error code
  repeat while (c := recvChar(ms)) <> "!"
  err := getHex(ms,2)                               ' Return code found
  checkNext(ms,13)
  
PUB checkErrorCode(ms, code) : e                    '' Make sure error code is zero
  if (e := getErrorCode(ms)) <> 0                   '  or equal to a specific code
    if e <> code
      abort ErrCode | e

PUB getErrorCode(ms) : err                          '' Receive an error code ("!00",13)
  checkNext(ms,"!")
  err := getHex(ms,2)
  checkNext(ms,13)

PUB checkNext(ms,char) | c                          '' Check the next received character
  if (c := recvChar(ms)) <> char
    abort (char << 8) | c

PUB getHex(ms,size) : hex | t                       '' Receive a fixed width hex value
  repeat size
    case t := recvChar(ms)
      "0".."9": hex := (hex << 4) | (t - "0")
      "A".."F": hex := (hex << 4) | (t - "A" + 10)
      "a".."f": hex := (hex << 4) | (t - "a" + 10)
      other: abort ("?" << 8) | t

PUB sendChar(ms,c) | time                           '' Send a character to the USBwiz
  if ms < 0
    case c                                          ' Echo to display if requested
      13, " ".."~":
        dsp.out(c)
  ms := (clkfreq / 1000) * ||ms
  time := cnt
  repeat while ina[USBwiz_Bsy] == 1                 ' Wait for USBwiz not Busy
    if (cnt - time) > ms
      abort TimeOut
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for previous I/O to finish
    if (cnt - time) > ms
      abort TimeOut
  long[ldr#ioControl+4] := (1 << 16) | @c
  long[ldr#ioControl] := (ldr#ioWrite0Cmd << 24) | USBwiz_Addr
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for this to finish
    if (cnt - time) > ms
      abort TimeOut
  if long[ldr#ioControl] & ldr#ioTestErr
    abort I2CIOErr

PUB sendStr(ms,s) | c                               '' Send a string to the USBwiz
  repeat while (c := byte[s++])
    sendChar(ms,c)

PUB sendHex(ms, v, w) | s, t                        '' Send a fixed width hex value
  repeat s from (w - 1) * 4 to 0 step 4
    if (t := ((v >> s) & $F)) < 10
      sendChar(ms, t + "0")                         ' "0".."9"
    else
      sendChar(ms, t - 10 + "A")                    ' "A".."F"
  
PUB recvChar(ms) | time, echo                       '' Receive a character from the USBwiz
  waitcnt(clkfreq / 100 + cnt)                      ' ***** Wait about 10ms between chars *****
  echo := ms < 0
  ms := (clkfreq / 1000) * ||ms
  time := cnt
  repeat until ina[USBwiz_Rdy] == 1                 ' See if character is present
    if (cnt - time) > ms
      abort TimeOut
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for previous I/O to finish
    if (cnt - time) > ms
      abort TimeOut
  long[ldr#ioControl+4] := (1 << 16) | @result
  long[ldr#ioControl] := (ldr#ioRead0Cmd << 24) | USBwiz_Addr | 1
  repeat while long[ldr#ioControl] & ldr#ioTestRdy  ' Wait for this to finish
    if (cnt - time) > ms
      abort TimeOut
  if long[ldr#ioControl] & ldr#ioTestErr
    abort I2CIOErr
  if echo                                           ' Echo to display if ms < 0
    case result
      13, " ".."~":
        dsp.out(result)
      other:
        dsp.out("<")
        dsp.dec(result)
        dsp.out(">")